@@ -5,8 +5,7 @@
 // SPDX-License-Identifier: GPL-3.0-or-later
 //
 
-#ifndef INCLUDED_UHD_USRP_MULTI_USRP_HPP
-#define INCLUDED_UHD_USRP_MULTI_USRP_HPP
+#pragma once
 
 // define API capabilities for compile time detection of new features
 #define UHD_USRP_MULTI_USRP_REF_SOURCES_API
@@ -23,9 +22,9 @@
 #define UHD_USRP_MULTI_USRP_FILTER_API
 #define UHD_USRP_MULTI_USRP_LO_CONFIG_API
 #define UHD_USRP_MULTI_USRP_TX_LO_CONFIG_API
+#define UHD_USRP_MULTI_USRP_POWER_LEVEL
 
 #include <uhd/config.hpp>
-#include <uhd/deprecated.hpp>
 #include <uhd/device.hpp>
 #include <uhd/types/filters.hpp>
 #include <uhd/types/ranges.hpp>
@@ -36,9 +35,10 @@
 #include <uhd/types/wb_iface.hpp>
 #include <uhd/usrp/dboard_iface.hpp>
 #include <uhd/usrp/subdev_spec.hpp>
-#include <boost/shared_ptr.hpp>
-#include <boost/utility.hpp>
+#include <uhd/utils/noncopyable.hpp>
+#include <uhd/rfnoc/radio_control.hpp>
 #include <complex>
+#include <memory>
 #include <string>
 #include <vector>
 
@@ -93,10 +93,10 @@ namespace usrp {
  *
  * </pre>
  */
-class UHD_API multi_usrp : boost::noncopyable
+class UHD_API multi_usrp : uhd::noncopyable
 {
 public:
-    typedef boost::shared_ptr<multi_usrp> sptr;
+    typedef std::shared_ptr<multi_usrp> sptr;
 
     virtual ~multi_usrp(void) = 0;
 
@@ -121,26 +121,25 @@ public:
      */
     static sptr make(const device_addr_t& dev_addr);
 
-    /*!
-     * Get the underlying device object.
-     * This is needed to get access to the streaming API and properties.
+    /*! Get the underlying device object
+     *
+     * Note that it is not recommended to use this method. The property tree can
+     * be accessed by calling get_tree() on this object, and the streamers own
+     * all the streaming-related functionality. get_tx_stream() and
+     * get_rx_stream() can also be called on this object.
+     *
+     * For RFNoC devices, this won't return a true uhd::device anyway, because
+     * direct device access is locked for those. The returned pointer will
+     * still point to a valid device object, however, it has reduced
+     * functionality.
+     *
      * \return the device object within this USRP
      */
     virtual device::sptr get_device(void) = 0;
 
-    /*! Returns true if this is a generation-3 device.
+    /*! Return a reference to the property tree
      */
-    virtual bool is_device3(void) = 0;
-
-    /*!
-     * Get the underlying device3 object. Only works for generation-3 (or later) devices.
-     *
-     * This is needed to get access to the streaming API and properties.
-     *
-     * \return The uhd::device3 object for this USRP.
-     * \throws uhd::type_error if this device is not actually a generation-3 device.
-     */
-    virtual boost::shared_ptr<uhd::device3> get_device3(void) = 0;
+    virtual uhd::property_tree::sptr get_tree(void) const = 0;
 
     //! Convenience method to get a RX streamer. See also uhd::device::get_rx_stream().
     virtual rx_streamer::sptr get_rx_stream(const stream_args_t& args) = 0;
@@ -213,7 +212,7 @@ public:
      * Examples:
      * - The B200 series' master clock rate can be changed at runtime and
      *   will report the true range of supported values
-     * - The X300 series has _two_ discrete options for the clock rate, but will
+     * - The X300 series has a valid range for the clock rate, but will
      *   always return the clock rate which the USRP was initialized to because
      *   it cannot be changed at runtime
      * - The N200 series does not have a configurable clock rate, and will
@@ -337,17 +336,6 @@ public:
     virtual void issue_stream_cmd(
         const stream_cmd_t& stream_cmd, size_t chan = ALL_CHANS) = 0;
 
-    /*!
-     * Set the clock configuration for the usrp device.
-     * DEPRECATED in favor of set time and clock source calls.
-     * This tells the usrp how to get a 10MHz reference and PPS clock.
-     * See the documentation for clock_config_t for more info.
-     * \param clock_config the clock configuration to set
-     * \param mboard which motherboard to set the config
-     */
-    virtual void set_clock_config(
-        const clock_config_t& clock_config, size_t mboard = ALL_MBOARDS) = 0;
-
     /*!  Set the time source for the USRP device
      *
      * This sets the method of time synchronization, typically a pulse per
@@ -568,13 +556,15 @@ public:
      * \param addr 8-bit register address
      * \param data 32-bit register value
      * \param mboard which motherboard to set the user register
+     * \throws uhd::not_implemented_error on RFNoC devices, uhd::lookup_error on
+     *         other devices if this API is not implemented.
      */
     virtual void set_user_register(
         const uint8_t addr, const uint32_t data, size_t mboard = ALL_MBOARDS) = 0;
 
     /*! Return a user settings interface object
      *
-     * This is only supported by some USRPs (B2xx series, N230). It will return
+     * This is only supported by the B2xx series. It will return
      * an object that will allow to peek and poke user settings, which typically
      * are implemented by custom FPGA images.
      * If the device does not support such an interface, it will return a null
@@ -598,6 +588,25 @@ public:
      */
     virtual uhd::wb_iface::sptr get_user_settings_iface(const size_t chan = 0) = 0;
 
+    /*! Get direct access to the underlying RFNoC radio object.
+     *
+     * Note: This is an advanced API, created for corner cases where the
+     * application is using multi_usrp, but some special features from
+     * radio_control need to be used that are not exposed by multi_usrp. Note
+     * that it is possible to put the radio and multi_usrp into a broken state
+     * by directly accessing the radio. For typical radio operations (such as
+     * tuning, setting gain or antenna, etc.) it is therefore highly recommended
+     * to not use this API call, but use the native multi_usrp API calls.
+     *
+     * The lifetime of the radio is linked to the lifetime of the device object,
+     * so storing a reference from this function is not allowed.
+     *
+     * \param chan The channel index
+     * \returns A reference to the radio block matching the given channel
+     * \throws uhd::not_implemented_error if not on an RFNoC device.
+     */
+    virtual uhd::rfnoc::radio_control& get_radio_control(const size_t chan = 0) = 0;
+
     /*******************************************************************
      * RX methods
      ******************************************************************/
@@ -640,6 +649,17 @@ public:
      */
     virtual void set_rx_rate(double rate, size_t chan = ALL_CHANS) = 0;
 
+    /*! Set the number of samples sent per packet (spp) for RX streaming
+     *
+     * On RFNoC devices, this will set the spp value on the radio itself. For
+     * older devices, it will inject the spp value into a later get_rx_stream()
+     * call, but it won't change anything in existing streamers.
+     *
+     * \param spp the new spp value
+     * \param chan the channel index 0 to N-1
+     */
+    virtual void set_rx_spp(const size_t spp, const size_t chan = ALL_CHANS) = 0;
+
     /*!
      * Gets the RX sample rate.
      * \param chan the channel index 0 to N-1
@@ -1179,6 +1199,70 @@ public:
     virtual void set_rx_iq_balance(
         const std::complex<double>& correction, size_t chan = ALL_CHANS) = 0;
 
+
+    /**************************************************************************
+     * Power level controls
+     *************************************************************************/
+    /*! Return true if this channel has a reference power API enabled
+     *
+     * Many devices either don't have a built-in reference power API, or they
+     * require calibration data for it to work. This means that it is not clear,
+     * even when the device type is known, if a device supports setting a power
+     * reference level. Use this method to query the availability of
+     * set_rx_power_reference() and get_rx_power_reference(), which will throw
+     * a uhd::not_implemented_error or uhd::runtime_error if they cannot be used.
+     *
+     * See \ref page_power for more information, or query the specific device's
+     * manual page to see if a power API is available, and how to enable it.
+     *
+     * \param chan The channel for which this feature is queried
+     *
+     * \returns true if this channel has an RX power API available
+     */
+    virtual bool has_rx_power_reference(const size_t chan = 0) = 0;
+
+    /*! Set the reference RX power level for a given channel
+     *
+     * Note: This functionality is not supported for most devices, and will
+     * cause a uhd::not_implemented_error exception to be thrown on devices that
+     * do not have this functionality.
+     *
+     * For more information on how to use this API, see \ref page_power.
+     *
+     * \param power_dbm The reference power level in dBm
+     * \param chan The channel for which this setting applies
+     *
+     * \throws uhd::not_implemented_error if this functionality does not exist
+     *         for this device
+     */
+    virtual void set_rx_power_reference(
+        const double power_dbm, const size_t chan = 0) = 0;
+
+    /*! Return the actual reference RX power level.
+     *
+     * Note: This functionality is not supported for most devices, and will
+     * cause a uhd::not_implemented_error exception to be thrown on devices that
+     * do not have this functionality.
+     *
+     * For more information on how to use this API, see \ref page_power.
+     *
+     * \param chan The channel for which this setting is queried
+     * \throws uhd::not_implemented_error if this functionality does not exist
+     *         for this device
+     */
+    virtual double get_rx_power_reference(const size_t chan = 0) = 0;
+
+    /*! Return the available RX power range given the current configuration
+     *
+     * This will return the range of available power levels given the current
+     * frequency, gain profile, antenna, and whatever other settings may affect
+     * the available power ranges. Note that the available power range may
+     * change frequently, so don't assume an immutable range.
+     *
+     * \param chan The channel index
+     */
+    virtual meta_range_t get_rx_power_range(const size_t chan) = 0;
+
     /*******************************************************************
      * TX methods
      ******************************************************************/
@@ -1375,6 +1459,69 @@ public:
      */
     virtual std::vector<std::string> get_tx_gain_names(size_t chan = 0) = 0;
 
+    /**************************************************************************
+     * Power level controls
+     *************************************************************************/
+    /*! Return true if this channel has a reference power API enabled
+     *
+     * Many devices either don't have a built-in reference power API, or they
+     * require calibration data for it to work. This means that it is not clear,
+     * even when the device type is known, if a device supports setting a power
+     * reference level. Use this method to query the availability of
+     * set_tx_power_reference() and get_tx_power_reference(), which will throw
+     * a uhd::not_implemented_error or uhd::runtime_error if they cannot be used.
+     *
+     * See \ref page_power for more information, or query the specific device's
+     * manual page to see if a power API is available, and how to enable it.
+     *
+     * \param chan The channel for which this feature is queried
+     *
+     * \returns true if this channel has a TX power API available
+     */
+    virtual bool has_tx_power_reference(const size_t chan = 0) = 0;
+
+    /*! Set the reference TX power level for a given channel
+     *
+     * Note: This functionality is not supported for most devices, and will
+     * cause a uhd::not_implemented_error exception to be thrown on devices that
+     * do not have this functionality.
+     *
+     * For more information on how to use this API, see \ref page_power.
+     *
+     * \param power_dbm The reference power level in dBm
+     * \param chan The channel for which this setting applies
+     *
+     * \throws uhd::not_implemented_error if this functionality does not exist
+     *         for this device
+     */
+    virtual void set_tx_power_reference(
+        const double power_dbm, const size_t chan = 0) = 0;
+
+    /*! Return the actual reference TX power level.
+     *
+     * Note: This functionality is not supported for most devices, and will
+     * cause a uhd::not_implemented_error exception to be thrown on devices that
+     * do not have this functionality.
+     *
+     * For more information on how to use this API, see \ref page_power.
+     *
+     * \param chan The channel for which this setting is queried
+     * \throws uhd::not_implemented_error if this functionality does not exist
+     *         for this device
+     */
+    virtual double get_tx_power_reference(const size_t chan = 0) = 0;
+
+    /*! Return the available TX power range given the current configuration
+     *
+     * This will return the range of available power levels given the current
+     * frequency, gain profile, antenna, and whatever other settings may affect
+     * the available power ranges. Note that the available power range may
+     * change frequently, so don't assume an immutable range.
+     *
+     * \param chan The channel index
+     */
+    virtual meta_range_t get_tx_power_range(const size_t chan) = 0;
+
     /*!
      * Select the TX antenna on the frontend.
      * \param ant the antenna name
@@ -1499,42 +1646,6 @@ public:
         const uint32_t mask = 0xffffffff,
         const size_t mboard = 0) = 0;
 
-    /*!
-     * Set a GPIO attribute on a particular GPIO bank.
-     * Possible attribute names:
-     *  - SRC  - "PS" for handling by processing system
-     *         - "RADIO_N/M" for handling by radio block with N is in [0..Number of
-     * Radio]; M is in [0..Number of port per Radio]
-     *  - CTRL - "ATR"  for ATR mode
-     *         - "GPIO" for GPIO mode
-     *  - DDR  - "OUT" for output
-     *         - "IN"  for input
-     *  - OUT -  a string of numbers representing GPIO output level (not ATR mode)
-     *        - "HIGH"or "LOW" as GPIO output level that apply for each bit mask that is 1
-     *  - ATR_0X - a string of numbers representing a value of the ATR idle state register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR idle state
-     * register
-     *  - ATR_RX - a string of numbers representing a value of a ATR receive only state
-     * register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR receive only
-     * state register
-     *  - ATR_TX - a string of numbers representing a value of the ATR transmit only state
-     * register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR transmit only
-     * state register
-     *  - ATR_XX - a string of numbers representing a value of the ATR full duplex state
-     * register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR full duplex
-     * state register \param bank the name of a GPIO bank \param attr the name of a GPIO
-     * attribute \param value the new value for this GPIO bank \param mask the bit mask to
-     * effect which pins are changed \param mboard the motherboard index 0 to M-1
-     */
-    virtual void set_gpio_attr(const std::string& bank,
-        const std::string& attr,
-        const std::string& value,
-        const uint32_t mask = 0xffffffff,
-        const size_t mboard = 0) = 0;
-
     /*!
      * Get a GPIO attribute on a particular GPIO bank.
      * Possible attribute names:
@@ -1554,126 +1665,125 @@ public:
     virtual uint32_t get_gpio_attr(
         const std::string& bank, const std::string& attr, const size_t mboard = 0) = 0;
 
-    /*!
-     * Get a GPIO attribute on a particular GPIO bank.
-     * Possible attribute names:
-     *  - SRC  - "PS" for handling by processing system
-     *         - "RADIO_N/M" for handling by radio block with N is in [0..Number of
-     * Radio]; M is in [0..Number of port per Radio]
-     *  - CTRL - "ATR"  for ATR mode
-     *         - "GPIO" for GPIO mode
-     *  - DDR  - "OUT" for output
-     *         - "IN"  for input
-     *  - OUT -  a string of numbers representing GPIO output level (not ATR mode)
-     *        - "HIGH"or "LOW" as GPIO output level that apply for each bit mask that is 1
-     *  - ATR_0X - a string of numbers representing a value of the ATR idle state register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR idle state
-     * register
-     *  - ATR_RX - a string of numbers representing a value of a ATR receive only state
-     * register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR receive only
-     * state register
-     *  - ATR_TX - a string of numbers representing a value of the ATR transmit only state
-     * register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR transmit only
-     * state register
-     *  - ATR_XX - a string of numbers representing a value of the ATR full duplex state
-     * register
-     *           - "HIGH" or "LOW" as a value set on each bit on of the ATR full duplex
-     * state register
-     *  - READBACK - readback input GPIOs
-     * \param bank the name of a GPIO bank
-     * \param attr the name of a GPIO attribute
+    /*! Return a list of GPIO banks that can be source-controlled on this motherboard
+     *
+     * This is a different set of banks than those returned from get_gpio_banks().
+     * Here, we return a list of banks that can be used as arguments for
+     * get_gpio_src(), get_gpio_srcs(), and set_gpio_src().
+     *
+     * Some motherboards have GPIO banks that can be driven from different
+     * sources, e.g., the N310 can have any radio channel drive the FP-GPIOs,
+     * or the PS.
+     *
      * \param mboard the motherboard index 0 to M-1
-     * \return the value set for this attribute in vector of strings
+     * \return a list of valid bank names
      */
-    virtual std::vector<std::string> get_gpio_string_attr(
-        const std::string& bank, const std::string& attr, const size_t mboard = 0) = 0;
+    virtual std::vector<std::string> get_gpio_src_banks(const size_t mboard = 0) = 0;
 
-    /*******************************************************************
-     * Register IO methods
-     ******************************************************************/
-    struct register_info_t
-    {
-        size_t bitwidth;
-        bool readable;
-        bool writable;
-    };
-
-    /*!
-     * Enumerate the full paths of all low-level USRP registers accessible to read/write
+    /*! Enumerate sources for a gpio bank on the specified device.
+     *
+     * Each of the pins in the chosen bank can be driven from one of the
+     * returned sources.
+     *
+     * \param bank the name of a GPIO bank. Valid values can be obtained by
+     *        calling get_gpio_src_banks().
      * \param mboard the motherboard index 0 to M-1
-     * \return a vector of register paths
+     * \return a list of strings with each valid source for the chosen bank
      */
-    virtual std::vector<std::string> enumerate_registers(const size_t mboard = 0) = 0;
+    virtual std::vector<std::string> get_gpio_srcs(
+        const std::string& bank, const size_t mboard = 0) = 0;
 
-    /*!
-     * Get more information about a low-level device register
-     * \param path the full path to the register
+    /*! Get the current source for each pin in a GPIO bank.
+     *
+     * \param bank the name of a GPIO bank. Valid values can be obtained by
+     *        calling get_gpio_src_banks().
      * \param mboard the motherboard index 0 to M-1
-     * \return the info struct which contains the bitwidth and read-write access
-     * information
+     * \return a list of strings for current source of each GPIO pin in the
+     *         chosen bank. The length of the return value matches the number of
+     *         programmable GPIO pins.
      */
-    virtual register_info_t get_register_info(
-        const std::string& path, const size_t mboard = 0) = 0;
+    virtual std::vector<std::string> get_gpio_src(
+        const std::string& bank, const size_t mboard = 0) = 0;
 
-    /*!
-     * Write a low-level register field for a register in the USRP hardware
-     * \param path the full path to the register
-     * \param field the identifier of bitfield to be written (all other bits remain
-     * unchanged) \param value the value to write to the register field \param mboard the
-     * motherboard index 0 to M-1
+    /*! Set the current source for each pin in a GPIO bank.
+     *
+     * Note: The length of the vector must be identical to the number of
+     * programmable GPIO pins.
+     *
+     * \param bank the name of a GPIO bank. Valid values can be obtained by
+     *        calling get_gpio_src_banks().
+     * \param src a list of strings specifying the source of each pin in a GPIO bank
+     * \param mboard the motherboard index 0 to M-1
      */
-    virtual void write_register(const std::string& path,
-        const uint32_t field,
-        const uint64_t value,
+    virtual void set_gpio_src(const std::string& bank,
+        const std::vector<std::string>& src,
         const size_t mboard = 0) = 0;
 
-    /*!
-     * Read a low-level register field from a register in the USRP hardware
-     * \param path the full path to the register
-     * \param field the identifier of bitfield to be read
-     * \param mboard the motherboard index 0 to M-1
-     * \return the value of the register field
-     */
-    virtual uint64_t read_register(
-        const std::string& path, const uint32_t field, const size_t mboard = 0) = 0;
-
     /*******************************************************************
      * Filter API methods
      ******************************************************************/
-
+    // TODO: This should be a const function, but I don't want to wrestle with the
+    // compiler right now
     /*!
-     * Enumerate the available filters in the signal path.
-     * \param search_mask
-     * \parblock
-     * Select only certain filter names by specifying this search mask.
-     *
-     * E.g. if search mask is set to "rx_frontends/A" only filter names including that
-     * string will be returned. \endparblock \return a vector of strings representing the
-     * selected filter names.
+     * Enumerate the available filters in the RX signal path.
+     * \param chan RX channel index 0 to N-1
+     * \return a vector of strings representing the selected filter names.
+     * \return Filter names will follow the pattern BLOCK_ID:FILTER_NAME. For example,
+     * "0/Radio#0:HB_0"
      */
-    virtual std::vector<std::string> get_filter_names(
-        const std::string& search_mask = "") = 0;
+    virtual std::vector<std::string> get_rx_filter_names(const size_t chan) = 0;
 
     /*!
-     * Return the filter object for the given name.
-     * \param path the name of the filter as returned from get_filter_names().
+     * Return the filter object for the given RX filter name.
+     * \param name the name of the filter as returned from get_rx_filter_names().
+     * \param chan RX channel index 0 to N-1
      * \return a filter_info_base::sptr.
      */
-    virtual filter_info_base::sptr get_filter(const std::string& path) = 0;
+    virtual uhd::filter_info_base::sptr get_rx_filter(
+        const std::string& name, const size_t chan) = 0;
 
     /*!
-     * Write back a filter obtained by get_filter() to the signal path.
+     * Write back a filter obtained by get_rx_filter() to the signal path.
      * This filter can be a modified version of the originally returned one.
-     * The information about Rx or Tx is contained in the path parameter.
-     * \param path the name of the filter as returned from get_filter_names().
+     * \param name the name of the filter as returned from get_rx_filter_names().
      * \param filter the filter_info_base::sptr of the filter object to be written
+     * \param chan RX channel index 0 to N-1
      */
-    virtual void set_filter(const std::string& path, filter_info_base::sptr filter) = 0;
+    virtual void set_rx_filter(const std::string& name,
+        uhd::filter_info_base::sptr filter,
+        const size_t chan) = 0;
+
+    // TODO: This should be a const function, but I don't want to wrestle with the
+    // compiler right now
+    /*!
+     * Enumerate the available filters in the TX signal path.
+     * \param chan TX channel index 0 to N-1
+     * \return a vector of strings representing the selected filter names.
+     * \return Filter names will follow the pattern BLOCK_ID:FILTER_NAME. For example,
+     * "0/Radio#0:HB_0"
+     */
+    virtual std::vector<std::string> get_tx_filter_names(const size_t chan) = 0;
+
+    /*!
+     * Return the filter object for the given TX filter name.
+     * \param name the name of the filter as returned from get_tx_filter_names().
+     * \param chan TX channel index 0 to N-1
+     * \return a filter_info_base::sptr.
+     */
+    virtual uhd::filter_info_base::sptr get_tx_filter(
+        const std::string& name, const size_t chan) = 0;
+
+    /*!
+     * Write back a filter obtained by get_tx_filter() to the signal path.
+     * This filter can be a modified version of the originally returned one.
+     * \param name the name of the filter as returned from get_tx_filter_names().
+     * \param filter the filter_info_base::sptr of the filter object to be written
+     * \param chan TX channel index 0 to N-1
+     */
+    virtual void set_tx_filter(const std::string& name,
+        uhd::filter_info_base::sptr filter,
+        const size_t chan) = 0;
 };
 
 } // namespace usrp
 } // namespace uhd
-
-#endif /* INCLUDED_UHD_USRP_MULTI_USRP_HPP */
